/* Programme de lecture d'une sonde de température/hygrométrie HDT11
*  Sauvegarde des données sur une carte SD
*  Configuration du programme par trame série
*  Code réalisé par Renan WAROUX pour le camps "Laboratoire du futur!"
*  Pour Objectif Sciences International
*  Date: 05/08/2011
*/

//Constantes, numéro de pin et variables
#include <SD.h>
#include <Wire.h>
#define NB_MAX_SAVE 60000 //Nombre maximal de sauvegarde par fichier
#define DHT11PIN 2
#define LEDPIN 13   // LED en PIN13 soudée sur l'Arduino        
#define PUMPPIN 12 //Le relai se trouve sur le pin 12
//#define DUREEARROSAGE 0x10 //Durée de l'arrosage en secondes
//#define PERIODEARROSAGE  3 //Période pendant laquelle on arrose pas, en heure
#define NB_INPUT_BYTES 35 //Nombre de byte en entrée

File myFile;

int cspin = 10; //Pin 10 pour le shield SD
int pin = 0; // analog pin
int tempc = 0; //variables de température
int samples[8]; // variables to make a better precision
int maxi = -100,mini = 100; // to start max/min temperature
int i;
long compteur = 0; //Un compteur pour compter le nombre de sauvegarde effectué
int indice_save = 1; //Un indice pour le numéro de la sauvegarde
char save_name[12]; //Nom de la sauvegarde
int humidity; //Variable pour stocker la valeur de l'humidité
int temperature; //Variable pour stocker la valeur de la température
int waiting_time = 5; //Temps d'attente en secondes

//pour l'arrosage
byte duree_arrosage = 0x10;//en seconde
byte periode_arrosage = 0x03; //en heure
//Message en provenance du pc
byte msg[NB_INPUT_BYTES];

//Pour l'horloge
byte secs,mins,hrs,jour,date,mois,annee;

//Fonction setup, s'exécute au démarrage
void setup()
{
  Serial.println("DHT11 READ PROGRAM ");
  Serial.println();
  
  //On attribue les pin pour le système d'arrosage
   pinMode(PUMPPIN, OUTPUT);
   digitalWrite(PUMPPIN, LOW); 
   pinMode(LEDPIN, OUTPUT);
   digitalWrite(LEDPIN, LOW);
 
  Wire.begin();//Initialisation de l'arduino en mode maitre
  Serial.begin(9600);//Début de la communication série
  //MiseALHeure(); //Mise à l'heure automatique, à décommenter pour mettre l'heure à jour

  Serial.print("Initializing SD card...");
  // On the Ethernet Shield, CS is pin 4. It's set as an output by default.
  // Note that even if it's not used as the CS pin, the hardware SS pin 
  // (10 on most Arduino boards, 53 on the Mega) must be left as an output 
  // or the SD library functions will not work. 
  pinMode(10, OUTPUT);
   
  if (!SD.begin(cspin))
    {
      Serial.println("initialization failed!");
      return;
    }
  
  Serial.println("initialization done.");
  
   //On sauvegarde les données
  // open the file. note that only one file can be open at a time,
  // so you have to close this one before opening another.
  
  // //On détermine le nom de la sauvegarde en fonction du numéro de la sauvegarde actuel
  if(indice_save > 0)
  {
    strcpy(save_name ,"save000.csv");
    if(indice_save < 9)
    {
      save_name[6] = indice_save+'0';
    }
    else if(indice_save < 99)
    {
      save_name[6] = indice_save%10 + '0';//On extrant les unités
      save_name[5] = (indice_save - indice_save%10)/10 + '0';//On extrait les dizaines
    }
    else if(indice_save < 999)
    {
      save_name[6] = indice_save%10 + '0';//On extrant les unités
      save_name[5] = (indice_save%100 - indice_save%10)/10 +'0';//On extrait les dizaines
      save_name[4] = (indice_save - indice_save%100)/100 +'0';//On extrait les centaines
    }
  }
  
  //On ouvre la sauvegarde
  myFile = SD.open(save_name, FILE_WRITE);
  
  // if the file opened okay, write to it:
  if (myFile) 
  {
    Serial.print("Writing to ");
    Serial.print(save_name);	
    // close the file:
    myFile.close();
    Serial.println(" done.");
  } 
  else
  {
    // if the file didn't open, print an error:
    Serial.print("error opening ");
    Serial.println(save_name);
    
  } 
}//Fin de la fonction setup

//Boucle principale, tourne en permanence
void loop()
{
  
  //Lecture sur le capteur T/H
  int chk = readDHT11(DHT11PIN);
  
  //Vérification des erreurs
  //Serial.print("Read sensor: ");
  switch (chk)
  {
    case 0: Serial.print("OK,"); break;
    case -1: Serial.print("Checksum error,"); break;
    case -2: Serial.print("Time out error,"); break;
    default: Serial.print("Unknown error,"); break;
  }
  
  //Lecture de données en provenance de l'ordinateur
  setData();
  
  //Horloge
  //Initialisation du pointeur de registre
  Wire.beginTransmission(0x68); //Condition de départ
  Wire.send(0); //Ecriture du pointeur de registre
  Wire.endTransmission();//Condition d'arret
  
  //On affiche les données
  affichage_donnees();
  
  //On lance la fonction d'arrosage
  arrosage();  
  //Sauvegarde des données
  writeToSDCard();  
  //On attends 5 s
  delay(waiting_time * 1000);

  tempc = 0;
  
  //On incrémente le compteur
  if( compteur < NB_MAX_SAVE)
  {
    compteur++;
  }
  else //Sinon, on remet le compteur à 0 et on incrémente le numéro de la sauvegarde
  {
    compteur = 0;
    indice_save++;
  }
}/* --(Fin de la boucle principale)-- */

/*-----( Declare User-written Functions )-----*/
//
//Celsius to Fahrenheit conversion
double Fahrenheit(double celsius)
{
	return 1.8 * celsius + 32;
}

//Celsius to Kelvin conversion
double Kelvin(double celsius)
{
	return celsius + 273.15;
}

// dewPoint function NOAA
// reference: http://wahiduddin.net/calc/density_algorithms.htm 
double dewPoint(double celsius, double humidity)
{
	double A0= 373.15/(273.15 + celsius);
	double SUM = -7.90298 * (A0-1);
	SUM += 5.02808 * log10(A0);
	SUM += -1.3816e-7 * (pow(10, (11.344*(1-1/A0)))-1) ;
	SUM += 8.1328e-3 * (pow(10,(-3.49149*(A0-1)))-1) ;
	SUM += log10(1013.246);
	double VP = pow(10, SUM-3) * humidity;
	double T = log(VP/0.61078);   // temp var
	return (241.88 * T) / (17.558-T);
}

// delta max = 0.6544 wrt dewPoint()
// 5x faster than dewPoint()
// reference: http://en.wikipedia.org/wiki/Dew_point
double dewPointFast(double celsius, double humidity)
{
	double a = 17.271;
	double b = 237.7;
	double temp = (a * celsius) / (b + celsius) + log(humidity/100);
	double Td = (b * temp) / (a - temp);
	return Td;
}

//Fonction de lecture du capteur DHT11
int readDHT11(int pin)
{
	// BUFFER TO RECEIVE
	uint8_t bits[5];
	uint8_t cnt = 7;
	uint8_t idx = 0;

	// EMPTY BUFFER
	for (int i=0; i< 5; i++) bits[i] = 0;

	// REQUEST SAMPLE
	pinMode(pin, OUTPUT);
	digitalWrite(pin, LOW);
	delay(18);
	digitalWrite(pin, HIGH);
	delayMicroseconds(40);
	pinMode(pin, INPUT);

	// ACKNOWLEDGE or TIMEOUT
	unsigned int loopCnt = 10000;
	while(digitalRead(pin) == LOW)
		if (loopCnt-- == 0) return -2;

	loopCnt = 10000;
	while(digitalRead(pin) == HIGH)
		if (loopCnt-- == 0) return -2;

	// READ OUTPUT - 40 BITS => 5 BYTES or TIMEOUT
	for (int i=0; i<40; i++)
	{
		loopCnt = 10000;
		while(digitalRead(pin) == LOW)
			if (loopCnt-- == 0) return -2;

		unsigned long t = micros();

		loopCnt = 10000;
		while(digitalRead(pin) == HIGH)
			if (loopCnt-- == 0) return -2;

		if ((micros() - t) > 40) bits[idx] |= (1 << cnt);
		if (cnt == 0)   // next byte?
		{
			cnt = 7;    // restart at MSB
			idx++;      // next byte!
		}
		else cnt--;
	}

	// WRITE TO RIGHT VARS
        // as bits[1] and bits[3] are allways zero they are omitted in formulas.
	humidity    = bits[0];
	temperature = bits[2];

	uint8_t sum = bits[0] + bits[2];

	if (bits[4] != sum) return -1;
	return 0;
}

//Fonction affichant les données sur la sortie série
void affichage_donnees()
{
  //On lit les données depuis l'horloge interne
  readFromClock();
  
  //Mise en forme et affichage des données
  //Mise en forme de l'heure
  if( hrs < 10) Serial.print("0");
  Serial.print(hrs,HEX);
  Serial.print(":");
  if( mins < 10) Serial.print("0");
  Serial.print(mins,HEX);
  Serial.print(":");
  if( secs < 10) Serial.print("0");
  Serial.print(secs,HEX);
  Serial.print(",");
  
  //Mise en forme de la date
  if( date < 10) Serial.print("0");
  Serial.print(date,HEX);
  Serial.print("/");
  
  if( mois < 10) Serial.print("0");
  Serial.print(mois, HEX);
  Serial.print("/");
  Serial.print("20"); //Ajout de 20 pour les années 20
  
  if(annee < 10 ) Serial.print("0");
  Serial.print(annee,HEX);
  
  //Affichage des données de température et d'humidité
  Serial.print(",");
  Serial.print((float)humidity, 2);

  Serial.print(",");
  Serial.print((float)temperature, 2);
  /*
  Serial.print("Temperature (oF): ");
  Serial.println(Fahrenheit(temperature), 2); */

  Serial.print(",");
  Serial.print(Kelvin(temperature), 2);
  
  Serial.print(",");
  Serial.print(dewPoint(temperature, humidity));
  
  //On affiche l'état de l'arroseur
  //1: Arroseur allumé, 0: arroseur éteint
  if( (hrs % periode_arrosage == 0) && (mins == 0x00) && (secs < duree_arrosage) ) //On arrose N seconde les N premières seconde d'une heure d'arrosage
  {
    Serial.print(",");
    Serial.println("1");
  }
  else
  {
    Serial.print(",");
    Serial.println("0");
  }
  /*
  Serial.print("Dew PointFast (oC): ");
  Serial.println(dewPointFast(temperature, humidity)); */
 
}

//Lire l'heure depuis l'horloge externe
void readFromClock()
{
   //Lecture des 7 octets contenant la date et l'heure
  Wire.requestFrom(0x68, 7);
  secs = Wire.receive();
  mins = Wire.receive();
  hrs = Wire.receive();
  jour = Wire.receive();
  date = Wire.receive();
  mois = Wire.receive();
  annee = Wire.receive();
  
}
 
//A décommenter pour mettre l'heure à jour
//Fonction de mise à l'heure simplifiée
/*void MiseALHeure()
{
  Wire.beginTransmission(0x68); //Condition de démarrage
  Wire.send(0); //Initialisation du pointeur de registres
  Wire.send(0x00); //Les secondes
  Wire.send(0x04); //Les minutes
  Wire.send(0x80 | 0x12); //L'heure au format 24H
  Wire.send(0x04);  //Jour de la semaine 01-07
  Wire.send(0x28); //Jour dans le mois 1-31
  Wire.send(0x07); //Mois de l'année 1-12
  Wire.send(0x11); //Année
  Wire.send(0x10); //Gestion de la sortie SQWOUT
  Wire.endTransmission(); //Condition d'arret
}*/

void getTemperature()
{
  //Lecture de la sonde de température
  for(i = 0;i <= 7;i++)// gets 8 samples of temperature
  { 
  samples[i] = ( 5.0 * analogRead(pin) * 100.0) / 1024.0;
  tempc = tempc + samples[i];
  }
  tempc = tempc/8.0; // better precision
  //tempf = (tempc * 9)/ 5 + 32; // converts to fahrenheit
  if(tempc > maxi) {maxi = tempc;} // set max temperature
  if(tempc < mini) {mini = tempc;} // set min temperature
    
}

//Cette fonction s'occupe d'ouvrir un fichier de sauvegarde pour sauver les mesures faites
void writeToSDCard()
{
  //On détermine le nom de la sauvegarde en fonction du numéro de la sauvegarde actuel
   if(indice_save > 0)
  {
    strcpy(save_name ,"save000.csv");
    if(indice_save < 9)
    {
      save_name[6] = indice_save+'0';
    }
    else if(indice_save < 99)
    {
      save_name[6] = indice_save%10 + '0';//On extrant les unités
      save_name[5] = (indice_save - indice_save%10)/10 + '0';//On extrait les dizaines
    }
    else if(indice_save < 999)
    {
      save_name[6] = indice_save%10 + '0';//On extrant les unités
      save_name[5] = (indice_save%100 - indice_save%10)/10 +'0';//On extrait les dizaines
      save_name[4] = (indice_save - indice_save%100)/100 +'0';//On extrait les centaines
    }
  }
  
  //On ouvre le fichier de sauvegarde
  myFile = SD.open(save_name, FILE_WRITE);
 
  // if the file opened okay, write to it:
  if (myFile) 
  {
    //Mise en forme et affichage des données
    if( hrs < 10) myFile.print("0");
    myFile.print(hrs,HEX);
    myFile.print(":");
    if( mins < 10) myFile.print("0");
    myFile.print(mins,HEX);
    myFile.print(":");
    if( secs < 10) myFile.print("0");
    myFile.print(secs,HEX);
    myFile.print(",");
    
    if( date < 10) myFile.print("0");
    myFile.print(date,HEX);
    myFile.print("/");
    
    if( mois < 10)myFile.print("0");
    myFile.print(mois, HEX);
    myFile.print("/");
    myFile.print("20"); //Ajout de 20 pour les années 20
    
    if(annee < 10 ) myFile.print("0");
    myFile.print(annee,HEX);
    
    //Sauvegarde des données de température et d'humidité
    myFile.print(",");
    myFile.print((float)humidity, 2);//Humidité

    myFile.print(",");
    myFile.print((float)temperature, 2);//Température en degrés

    myFile.print(", ");
    myFile.print(Kelvin(temperature), 2);//température en Kelvin
  
    myFile.print(",");
    myFile.print(dewPoint(temperature, humidity));//Point de rosée
    
      //On affiche l'état de l'arroseur
    //1: Arroseur allumé, 0: arroseur éteint
    if( (hrs % periode_arrosage == 0) && (mins == 0x00) && (secs < duree_arrosage )) //On arrose N seconde les N premières seconde d'une heure d'arrosage
    {
      myFile.print(",");
      myFile.println("1");
    }
    else
    {
      myFile.print(",");
      myFile.println("0");
    }
	
    //Fermeture du fichier:
    myFile.close();
  } 
  else
  {
    // if the file didn't open, print an error:
    Serial.print("error opening ");
    Serial.println(save_name);
  }  
}

void arrosage()
{
  
  if( (hrs % periode_arrosage == 0) && (mins == 0x00) && (secs < duree_arrosage) ) //On arrose N seconde les N premières seconde d'une heure d'arrosage
  {
    // Actionne la pompe et la LED
    digitalWrite(PUMPPIN, HIGH);
    digitalWrite(LEDPIN, HIGH);
   // Serial.println("Arrosage en cours");
  }
  else
  {
    //On désactive la pompe
    digitalWrite(PUMPPIN, LOW);
    digitalWrite(LEDPIN, LOW);
  }
}

//Cette fonction permet de traiter une trame série reçu par le µP
//De la forme: 18:09:00,03,22/08/2011,02,987,03,10
//HH:MM:SS,DD,JJ/MM/AA,XX,XXX,XX,XX
//Heure,Jour de la semaine,Date,Temps d'attente avant mesure,indice de la sauvegarde,heure d'attentes avant arrosage, temps arrosage (s)
void setData()
{
    // check if data has been sent from the computer:
  if (Serial.available()) {
    // read the most recent byte (which will be from 0 to 255):
    for(i=0; i < NB_INPUT_BYTES; i++)
    {
      msg[i] = Serial.read();
      Serial.print(msg[i]);
    }
    Serial.println();
    
    //On met à jour l'heure et la date de l'horloge
    hrs = ((msg[0] - '0')*16 + (msg[1] - '0'));//Pour l'heure
    mins = ((msg[3] - '0')*16 + (msg[4] - '0'));//Pour les minutes
    secs = ((msg[6] - '0')*16 + (msg[7] - '0'));//Pour les secondes
    jour = msg[9];//Pour l'heure
    date = ((msg[12] - '0')*16 + (msg[13] - '0'));//Pour la date
    mois = ((msg[15] - '0')*16 + (msg[16] - '0'));//Pour le mois
    annee = ((msg[20] - '0')*16 + (msg[21] - '0'));//Pour l'année
    
    //On applique les changement
    Wire.beginTransmission(0x68); //Condition de démarrage
    Wire.send(0); //Initialisation du pointeur de registres
    Wire.send(secs); //Les seconde
    Wire.send(mins); //Les minutes
    Wire.send(0x80 | hrs); //L'heure au format 24H
    Wire.send(jour);  //Jour de la semaine 01-07
    Wire.send(date); //Jour dans le mois 1-31
    Wire.send(mois); //Mois de l'année 1-12
    Wire.send(annee);//Année
    Wire.send(0x10); //Gestion de la sortie SQWOUT
    Wire.endTransmission(); //Condition d'arret
      
    //On set le temps d'attente entre chaque mesure
    waiting_time = (msg[23] - '0')*10 + ( msg[24] - '0');//en secondes
    
    //indice de la sauvegarde
    indice_save = (msg[26] - '0')*100 + (msg[27] - '0')*10 + ( msg[28] - '0');//en secondes
    
    //Paramètres de l'arrosage
    periode_arrosage = ((msg[30] - '0')*16 + ( msg[31] - '0')); //en heure
    duree_arrosage = ((msg[33] - '0')*16 + ( msg[34] - '0'));//en secondes


    Serial.print("Nouveau temps d'attente: ");
    Serial.println(waiting_time);
    Serial.print("Nouvel indice de sauvegarde: ");
    Serial.println(indice_save);
    Serial.print("Période d'arrosage: ");
    Serial.println(periode_arrosage,HEX);
    Serial.print("Duree arrosage: ");
    Serial.println(duree_arrosage,HEX);
  }
}
